import asyncio
import binascii
import fcntl
import logging
import os
import re
import subprocess

from shutil import which

from .base import InternalPassiveClient, InternalPassiveException, InvalidModbusFrame


class InvalidIpv4Packet(InternalPassiveException):
    pass


class InternalTcpDumpClient(InternalPassiveClient):
    _WORD_HEX_REGEX = re.compile(r"\s[a-f0-9]{2,4}")

    def __init__(
        self,
        *,
        port: int,
        iface: str | None = None,
        unit_id: int = 0,
        master_ip: str | None = None,
        slave_ip: str | None = None,
        master_port: int | None = None,
        slave_port: int | None = None,
        tcpdump: str | None = None,
        logger: logging.Logger,
    ) -> None:
        self.port: int = port
        self.iface: str = iface or "any"
        self.unit_id: int = unit_id
        self.master_ip: str | None = master_ip
        self.slave_ip: str | None = slave_ip
        self.master_port: int | None = master_port
        self.slave_port: int | None = slave_port
        self.tcpdump: str | None = tcpdump
        self.logger = logger

        if not any((self.master_ip, self.slave_ip, self.master_port, self.slave_port)):
            raise ValueError("Either `master_ip`, `slave_ip`, `master_port` or `slave_port` must be set")

        self.buffer: bytes = b""

        super().__init__(logger)
        self.set_logger_name("passive_tcpdump")

    @staticmethod
    def find_tcpdump() -> str | None:
        return which("tcpdump")

    def make_tcpdump_cmd(self) -> list[str]:
        tcpdump = self.tcpdump or self.find_tcpdump()
        if not tcpdump:
            raise RuntimeError("tcpdump command not found")

        return [tcpdump, "-Xni", self.iface, "port", str(self.port)]

    async def run(self) -> None:
        cmd = self.make_tcpdump_cmd()
        self.debug('running command "%s"', " ".join(cmd))
        with subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
        ) as proc:
            # Set non-blocking read mode
            fd = proc.stdout.fileno()
            flags = fcntl.fcntl(fd, fcntl.F_GETFL)
            fcntl.fcntl(fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)
            line = ""
            try:
                while True:
                    exit_code = proc.poll()
                    if exit_code is not None:
                        self.debug("tcpdump exited with code %s", exit_code)
                        break
                    try:
                        output = proc.stdout.read()
                    except TypeError:
                        # Sometimes the `read()` implementation raises `TypeError: can't concat NoneType to bytes`
                        output = None
                    if not output:
                        await asyncio.sleep(0.1)
                        continue
                    parts = output.split("\n")
                    line += parts.pop(0)
                    self._parse_line(line)
                    for line in parts[:-1]:
                        if line:
                            self._parse_line(line)
                    line = parts[-1]
            except asyncio.CancelledError:
                self.debug("stopping tcpdump process")
                proc.kill()

    def _parse_line(self, line: str) -> None:
        # Each `line` starts with a timestamp or contains something like:
        #     [tab]0x0080:  5375 6e2c 2031 3020 4465 6320 3230 3233  Sun,.10.Dec.2023
        if line.startswith("\t"):
            for word_hex in self._WORD_HEX_REGEX.findall(line):
                # Our `word_hex` looks like: ` 6e2c`, containing two bytes. We push each byte in network byte order.
                if len(word_hex) == 5:
                    self.buffer += bytes([int(word_hex[:3], 16), int(word_hex[3:], 16)])
                else:
                    self.buffer += bytes([int(word_hex[:3], 16)])
        elif self.buffer:
            try:
                self._process_packet(self.buffer)
            except Exception:
                self.error("failed to parse IP packet", exc_info=True)

            self.buffer = b""

    def _process_packet(self, data: bytes) -> None:
        # Validate IPv4 header
        if len(data) < 40:
            raise InvalidIpv4Packet("Packet too short (%s bytes)", len(data))
        if data[0] >> 4 != 4:
            raise InvalidIpv4Packet("Wrong IPv4 version (%s)", data[0] >> 4)

        ip_packet = data
        ip_header_length = (ip_packet[0] & 0x0F) * 4
        ip_source = f"{ip_packet[12]}.{ip_packet[13]}.{ip_packet[14]}.{ip_packet[15]}"
        ip_dest = f"{ip_packet[16]}.{ip_packet[17]}.{ip_packet[18]}.{ip_packet[19]}"

        tcp_packet = ip_packet[ip_header_length:]
        tcp_source_port = (tcp_packet[0] << 8) + tcp_packet[1]
        tcp_dest_port = (tcp_packet[2] << 8) + tcp_packet[3]
        tcp_header_length = (tcp_packet[12] >> 4) * 4

        # if self.port not in (tcp_source_port, tcp_dest_port):
        #     self.debug('ignoring message on port %s', self.port)
        #     return

        modbus_frame = tcp_packet[tcp_header_length:]
        if not modbus_frame:
            # Ignore packet with empty TCP data
            return

        is_request = (
            (self.master_ip and self.master_ip == ip_source)
            or (self.slave_ip and self.slave_ip == ip_dest)
            or (self.master_port and self.master_port == tcp_source_port)
            or (self.slave_port and self.slave_port == tcp_dest_port)
        ) or False
        is_response = (
            (self.master_ip and self.master_ip == ip_dest)
            or (self.slave_ip and self.slave_ip == ip_source)
            or (self.master_port and self.master_port == tcp_dest_port)
            or (self.slave_port and self.slave_port == tcp_source_port)
        ) or False
        if is_request:
            self._process_modbus_request(modbus_frame)
        elif is_response:
            self._process_modbus_response(modbus_frame)

    def _process_modbus_request(self, frame: bytes) -> None:
        self.debug("got Modbus request: %s", binascii.hexlify(frame).decode())
        if len(frame) < 5:
            raise InvalidModbusFrame("Frame too short (%s bytes)", len(frame))

        unit_id = frame[0]
        function = frame[1]
        data = frame[2:-2]
        crc = frame[-2:]
        if crc != self.compute_crc(frame[:-2]):
            raise InvalidModbusFrame("CRC verification failed")
        if function not in self.WATCHED_FUNCTIONS:
            self.debug("ignoring Modbus function %s", function)
            return
        if self.unit_id and self.unit_id != unit_id:
            self.debug("ignoring Modbus slave address %s", unit_id)
            return
        if function == self.MODBUS_FUNC_READ_COILS:
            self.process_read_coils_request(data)
        elif function == self.MODBUS_FUNC_READ_DISCRETE_INPUTS:
            self.process_read_discrete_inputs_request(data)
        elif function == self.MODBUS_FUNC_READ_HOLDING_REGISTERS:
            self.process_read_holding_registers_request(data)
        elif function == self.MODBUS_FUNC_READ_INPUT_REGISTERS:
            self.process_read_input_registers_request(data)

    def _process_modbus_response(self, frame: bytes) -> None:
        self.debug("got Modbus response: %s", binascii.hexlify(frame).decode())
        if len(frame) < 5:
            raise InvalidModbusFrame("Frame too short (%s bytes)", len(frame))

        unit_id = frame[0]
        function = frame[1]
        data = frame[2:-2]
        crc = frame[-2:]
        if crc != self.compute_crc(frame[:-2]):
            raise InvalidModbusFrame("CRC verification failed")
        if function not in self.WATCHED_FUNCTIONS:
            self.debug("ignoring Modbus function %s", function)
            return
        if self.unit_id and self.unit_id != unit_id:
            self.debug("ignoring Modbus unit ID %s", unit_id)
            return
        if function == self.MODBUS_FUNC_READ_COILS:
            self.process_read_coils_response(data)
        elif function == self.MODBUS_FUNC_READ_DISCRETE_INPUTS:
            self.process_read_discrete_inputs_response(data)
        elif function == self.MODBUS_FUNC_READ_HOLDING_REGISTERS:
            self.process_read_holding_registers_response(data)
        elif function == self.MODBUS_FUNC_READ_INPUT_REGISTERS:
            self.process_read_input_registers_response(data)



from .client import ModbusSerialClient, ModbusTcpClient, PassiveModbusSerialClient, PassiveModbusTcpClient


__all__ = ["ModbusSerialClient", "ModbusTcpClient", "PassiveModbusSerialClient", "PassiveModbusTcpClient"]


VERSION = "0.0.0"



import abc
import asyncio
import logging

from typing import Any

from pymodbus.client import AsyncModbusSerialClient as InternalAsyncModbusSerialClient
from pymodbus.client import AsyncModbusTcpClient as InternalAsyncModbusTcpClient
from pymodbus.client.base import ModbusBaseClient as InternalModbusBaseClient
from pymodbus.pdu import ExceptionResponse
from qtoggleserver.core import ports as core_ports
from qtoggleserver.lib import polled
from qtoggleserver.utils import json as json_utils

from . import constants
from .base import BaseModbus
from .passive.serial import InternalSerialClient
from .passive.tcpdump import InternalTcpDumpClient


class BaseModbusClient(polled.PolledPeripheral, BaseModbus, metaclass=abc.ABCMeta):
    DEFAULT_POLL_INTERVAL = 5

    logger = logging.getLogger(__name__)

    def __init__(self, initial_delay: int = 0, **kwargs) -> None:
        polled.PolledPeripheral.__init__(self, **kwargs)
        BaseModbus.__init__(self, **kwargs)

        self.initial_delay: int = initial_delay

        self._values_by_type_and_address: dict[str, dict[int, Any]] = {}
        self._lengths_by_type_and_address: dict[str, dict[int, int]] = {}
        for port_detail in self.port_details.values():
            address = port_detail["address"]
            # Convert string numbers (e.g. hex) to integer
            if isinstance(address, str):
                address = int(address, base=0)
            lengths_by_address = self._lengths_by_type_and_address.setdefault(port_detail["modbus_type"], {})
            length = port_detail.get("length", 1)
            lengths_by_address[address] = length

        # Merge intersecting address spaces
        for modbus_type, lengths_by_address in self._lengths_by_type_and_address.items():
            lengths_by_address_items = list(lengths_by_address.items())
            i = 0
            while i < len(lengths_by_address_items) - 1:
                j = i + 1
                while j < len(lengths_by_address_items):
                    address1, length1 = lengths_by_address_items[i]
                    address2, length2 = lengths_by_address_items[j]
                    merged = self._try_merge_address_length(address1, length1, address2, length2)
                    if merged:
                        lengths_by_address_items[i] = merged
                        lengths_by_address_items.pop(j)
                    else:
                        j += 1
                i += 1

            self._lengths_by_type_and_address[modbus_type] = dict(lengths_by_address_items)

        self._pymodbus_client: InternalModbusBaseClient | None = None

    @abc.abstractmethod
    async def make_internal_client(self) -> InternalModbusBaseClient:
        raise NotImplementedError()

    @staticmethod
    def _try_merge_address_length(address1: int, length1: int, address2: int, length2: int) -> tuple[int, int] | None:
        end1 = address1 + length1
        end2 = address2 + length2
        if address1 <= address2:
            if end1 >= address2:
                if end1 <= end2:  # address1, address2, end1, end2
                    return address1, end2 - address1
                else:  # address1, address2, end2, end1
                    return address1, end1 - address1
        else:  # address1 > address2
            if end2 >= address1:
                if end2 <= end1:  # address2, address1, end2, end1
                    return address2, end1 - address2
                else:  # address2, address1, end1, end2
                    return address2, end2 - address2

        return None

    async def ensure_client(self) -> bool:
        if self._pymodbus_client and self._pymodbus_client.connected:
            return True

        self.info("connecting to unit")
        self._pymodbus_client = await self.make_internal_client()
        await self._pymodbus_client.connect()

        if self.initial_delay:
            self.debug("waiting %d seconds for initial delay", self.initial_delay)
            await asyncio.sleep(self.initial_delay)

        for _ in range(self.timeout * 10):
            if self._pymodbus_client.connected:
                break
            await asyncio.sleep(0.1)

        if not self._pymodbus_client.connected:
            self.warning("could not connect to unit")
            # Don't leave the client (partially) connected
            try:
                self._pymodbus_client.close()
            except Exception:
                # We don't care if connection closing fails - we're going to recreate the client from scratch anyway
                pass
            return False

        return True

    def close_client(self) -> None:
        self.info("disconnecting from unit")
        try:
            self._pymodbus_client.close()
        except Exception:
            # We don't care if connection closing fails - we're going to recreate the client from scratch anyway
            pass

        self._pymodbus_client = None

    async def make_port_args(self) -> list[dict[str, Any] | type[core_ports.BasePort]]:
        from .ports import ModbusClientPort

        port_args = []
        for id_, details in self.port_details.items():
            port_args.append({"driver": ModbusClientPort, "id": id_, **details})

        return port_args

    async def handle_enable(self) -> None:
        await self.ensure_client()
        await super().handle_enable()

    async def handle_disable(self) -> None:
        await super().handle_disable()
        self.close_client()

    async def poll(self) -> None:
        if not await self.ensure_client():
            raise Exception("Could not connect to Modbus unit")

        values_by_type_and_address: dict[str, dict[int, Any]] = {}
        for modbus_type, lengths_by_address in self._lengths_by_type_and_address.items():
            for address, length in lengths_by_address.items():
                if modbus_type == constants.MODBUS_TYPE_COIL:
                    self.debug("reading %d coils at 0x%04X", length, address)
                    response = await self._pymodbus_client.read_coils(address, count=length, slave=self.unit_id)
                    if isinstance(response, ExceptionResponse):
                        raise Exception(f"Got Modbus erroneous response: {response}")

                    values = response.bits
                    values_str = ", ".join(str(v) for v in values).lower()
                    self.debug("read coil values (%s) at 0x%04X", values_str, address)
                elif modbus_type == constants.MODBUS_TYPE_DISCRETE_INPUT:
                    self.debug("reading %d discrete inputs at 0x%04X", length, address)
                    response = await self._pymodbus_client.read_discrete_inputs(
                        address, count=length, slave=self.unit_id
                    )
                    if isinstance(response, ExceptionResponse):
                        raise Exception(f"Got Modbus erroneous response: {response}")

                    self.error("got erroneous response: %s", str(response))
                    values = response.bits
                    values_str = ", ".join(str(v) for v in values).lower()
                    self.debug("read discrete input values (%s) at 0x%04X", values_str, address)
                elif modbus_type == constants.MODBUS_TYPE_INPUT_REGISTER:
                    self.debug("reading %d input registers at 0x%04X", length, address)
                    response = await self._pymodbus_client.read_input_registers(
                        address, count=length, slave=self.unit_id
                    )
                    if isinstance(response, ExceptionResponse):
                        raise Exception(f"Got Modbus erroneous response: {response}")

                    values = response.registers
                    values_str = ", ".join(str(v) for v in values)
                    self.debug("read input registers values (%s) at 0x%04X", values_str, address)
                elif modbus_type == constants.MODBUS_TYPE_HOLDING_REGISTER:
                    self.debug("reading %d holding registers at 0x%04X", length, address)
                    response = await self._pymodbus_client.read_holding_registers(
                        address, count=length, slave=self.unit_id
                    )
                    if isinstance(response, ExceptionResponse):
                        raise Exception(f"Got Modbus erroneous response: {response}")

                    values = response.registers
                    values_str = ", ".join(str(v) for v in values)
                    self.debug("read holding registers values (%s) at 0x%04X", values_str, address)
                else:
                    continue

                if len(values) < length:
                    raise Exception("Unexpected number of values read: %s < %s", len(values), length)

                for i in range(length):
                    values_by_type_and_address.setdefault(modbus_type, {})[address + i] = values[i]

            self._values_by_type_and_address = values_by_type_and_address

    def get_last_coil_value(self, address: int) -> bool | None:
        values_by_address = self._values_by_type_and_address.get(constants.MODBUS_TYPE_COIL, {})
        return values_by_address.get(address)

    def get_last_discrete_input_value(self, address: int) -> bool | None:
        values_by_address = self._values_by_type_and_address.get(constants.MODBUS_TYPE_DISCRETE_INPUT, {})
        return values_by_address.get(address)

    def get_last_input_register_values(self, address: int, length: int) -> list[int] | None:
        values_by_address = self._values_by_type_and_address.get(constants.MODBUS_TYPE_INPUT_REGISTER, {})
        values = []
        for i in range(length):
            value = values_by_address.get(address + i)
            if value is None:  # a single None value in the list makes the entire result None
                return
            values.append(value)

        return values

    def get_last_holding_register_values(self, address: int, length: int) -> list[int] | None:
        values_by_address = self._values_by_type_and_address.get(constants.MODBUS_TYPE_HOLDING_REGISTER, {})
        values = []
        for i in range(length):
            value = values_by_address.get(address + i)
            if value is None:  # a single None value in the list makes the entire result None
                return
            values.append(value)

        return values

    async def write_coil_value(self, address: int, value: bool) -> None:
        self.debug("writing coil value %s to 0x%04X", json_utils.dumps(value), address)
        if self.use_single_functions:
            await self._pymodbus_client.write_coil(address, value, slave=self.unit_id)
        else:
            await self._pymodbus_client.write_coils(address, [value], slave=self.unit_id)

        self._values_by_type_and_address.setdefault(constants.MODBUS_TYPE_COIL, {})[address] = value

    async def write_holding_register_values(self, address: int, values: list[int]) -> None:
        values_str = ", ".join(["%04X" % v for v in values])
        self.debug("writing holding register values %s to 0x%04X", values_str, address)
        if self.use_single_functions:
            for value in values:
                await self._pymodbus_client.write_register(address, value, slave=self.unit_id)
        else:
            await self._pymodbus_client.write_registers(address, values, slave=self.unit_id)

        for i, value in enumerate(values):
            self._values_by_type_and_address.setdefault(constants.MODBUS_TYPE_HOLDING_REGISTER, {})[address + i] = value


class ModbusSerialClient(BaseModbusClient):
    DEFAULT_METHOD = "ascii"
    DEFAULT_SERIAL_BAUD = 9600
    DEFAULT_SERIAL_STOPBITS = 1
    DEFAULT_SERIAL_BYTESIZE = 8
    DEFAULT_SERIAL_PARITY = "N"

    def __init__(
        self,
        *,
        serial_port: str,
        serial_baud: int = DEFAULT_SERIAL_BAUD,
        serial_stopbits: int = DEFAULT_SERIAL_STOPBITS,
        serial_bytesize: int = DEFAULT_SERIAL_BYTESIZE,
        serial_parity: str = DEFAULT_SERIAL_PARITY,
        **kwargs,
    ) -> None:
        self.serial_port: str = serial_port
        self.serial_baud: int = serial_baud
        self.serial_stopbits: int = serial_stopbits
        self.serial_bytesize: int = serial_bytesize
        self.serial_parity: str = serial_parity

        kwargs.setdefault("method", self.DEFAULT_METHOD)

        super().__init__(**kwargs)

    async def make_internal_client(self) -> InternalModbusBaseClient:
        framer_cls = self.FRAMERS_BY_METHOD[self.method]

        return InternalAsyncModbusSerialClient(
            self.serial_port,
            baudrate=self.serial_baud,
            stopbits=self.serial_stopbits,
            bytesize=self.serial_bytesize,
            parity=self.serial_parity,
            timeout=self.timeout,
            framer=framer_cls,
        )


class ModbusTcpClient(BaseModbusClient):
    DEFAULT_METHOD = "socket"
    DEFAULT_TCP_PORT = 502

    def __init__(self, *, tcp_host: str, tcp_port: int = DEFAULT_TCP_PORT, **kwargs) -> None:
        self.tcp_host: str = tcp_host
        self.tcp_port: int = tcp_port

        kwargs.setdefault("method", self.DEFAULT_METHOD)
        super().__init__(**kwargs)

    async def make_internal_client(self) -> InternalModbusBaseClient:
        framer_cls = self.FRAMERS_BY_METHOD[self.method]

        return InternalAsyncModbusTcpClient(
            host=self.tcp_host,
            port=self.tcp_port,
            framer=framer_cls,
            timeout=self.timeout,
            reconnect_delay=0,
        )

    def handle_offline(self) -> None:
        # Automatically close and cleanup any existing client if peripheral goes offline. We want to start from scratch
        # with a brand-new client.
        self.close_client()


class BasePassiveModbusClient(BaseModbusClient, metaclass=abc.ABCMeta):
    def __init__(self, **kwargs) -> None:
        super().__init__(method="ignored", **kwargs)


class PassiveModbusSerialClient(BasePassiveModbusClient):
    DEFAULT_SERIAL_BAUD = 9600
    DEFAULT_SERIAL_STOPBITS = 1
    DEFAULT_SERIAL_BYTESIZE = 8
    DEFAULT_SERIAL_PARITY = "N"

    def __init__(
        self,
        *,
        serial_port: str,
        serial_baud: int = DEFAULT_SERIAL_BAUD,
        serial_stopbits: int = DEFAULT_SERIAL_STOPBITS,
        serial_bytesize: int = DEFAULT_SERIAL_BYTESIZE,
        serial_parity: str = DEFAULT_SERIAL_PARITY,
        **kwargs,
    ) -> None:
        self.serial_port: str = serial_port
        self.serial_baud: int = serial_baud
        self.serial_stopbits: int = serial_stopbits
        self.serial_bytesize: int = serial_bytesize
        self.serial_parity: str = serial_parity

        super().__init__(**kwargs)

    async def make_internal_client(self) -> InternalModbusBaseClient:
        return InternalSerialClient(
            serial_port=self.serial_port,
            serial_baud=self.serial_baud,
            serial_stopbits=self.serial_stopbits,
            serial_bytesize=self.serial_bytesize,
            serial_parity=self.serial_parity,
            unit_id=self.unit_id,
            timeout=self.timeout,
            logger=self.logger,
        )


class PassiveModbusTcpClient(BasePassiveModbusClient):
    def __init__(
        self,
        *,
        port: int,
        iface: str = "any",
        master_ip: str | None = None,
        slave_ip: str | None = None,
        master_port: int | None = None,
        slave_port: int | None = None,
        tcpdump: str | None = None,
        **kwargs,
    ) -> None:
        self.port: int = port
        self.iface: str = iface
        self.master_ip: str | None = master_ip
        self.slave_ip: str | None = slave_ip
        self.master_port: int | None = master_port
        self.slave_port: int | None = slave_port
        self.tcpdump: str | None = tcpdump

        super().__init__(**kwargs)

    async def make_internal_client(self) -> InternalModbusBaseClient:
        return InternalTcpDumpClient(
            port=self.port,
            iface=self.iface,
            unit_id=self.unit_id,
            master_ip=self.master_ip,
            slave_ip=self.slave_ip,
            master_port=self.master_port,
            slave_port=self.slave_port,
            tcpdump=self.tcpdump,
            logger=self.logger,
        )



import logging

from typing import Any

from pymodbus.framer.ascii_framer import ModbusAsciiFramer
from pymodbus.framer.binary_framer import ModbusBinaryFramer
from pymodbus.framer.rtu_framer import ModbusRtuFramer
from pymodbus.framer.socket_framer import ModbusSocketFramer


class BaseModbus:
    DEFAULT_TIMEOUT = 10  # seconds
    DEFAULT_UNIT_ID = 0

    FRAMERS_BY_METHOD = {
        "ascii": ModbusAsciiFramer,
        "rtu": ModbusRtuFramer,
        "binary": ModbusBinaryFramer,
        "socket": ModbusSocketFramer,
    }

    def __init__(
        self,
        *,
        method: str,
        unit_id: int = DEFAULT_UNIT_ID,
        timeout: int = DEFAULT_TIMEOUT,
        use_single_functions: bool = False,
        ports: dict[str, dict[str, Any]],
        **kwargs,
    ) -> None:
        self.method: str = method
        self.unit_id: int = unit_id
        self.timeout: int = timeout
        self.use_single_functions: bool = use_single_functions
        self.port_details: dict[str, dict[str, Any]] = ports

        # TODO: use a common way of adjusting default logging settings from addons
        logging.getLogger("pymodbus").setLevel(logging.INFO)



import abc
import asyncio
import logging
import math

from typing import Any

from pymodbus import Framer, bit_read_message, bit_write_message, register_read_message, register_write_message
from pymodbus.client.base import ModbusBaseClient as InternalModbusBaseClient
from qtoggleserver.utils import logging as logging_utils


class InternalPassiveException(Exception):
    pass


class InvalidModbusFrame(InternalPassiveException):
    pass


class InternalPassiveClient(InternalModbusBaseClient, logging_utils.LoggableMixin, metaclass=abc.ABCMeta):
    MODBUS_FUNC_READ_COILS = 0x01
    MODBUS_FUNC_READ_DISCRETE_INPUTS = 0x02
    MODBUS_FUNC_READ_HOLDING_REGISTERS = 0x03
    MODBUS_FUNC_READ_INPUT_REGISTERS = 0x04
    MODBUS_FUNC_WRITE_SINGLE_COIL = 0x05
    MODBUS_FUNC_WRITE_SINGLE_REGISTER = 0x6
    MODBUS_FUNC_WRITE_MULTIPLE_COILS = 0x15
    MODBUS_FUNC_WRITE_MULTIPLE_REGISTERS = 0x16

    ALL_FUNCTIONS = {
        MODBUS_FUNC_READ_COILS,
        MODBUS_FUNC_READ_DISCRETE_INPUTS,
        MODBUS_FUNC_READ_HOLDING_REGISTERS,
        MODBUS_FUNC_READ_INPUT_REGISTERS,
        MODBUS_FUNC_WRITE_SINGLE_COIL,
        MODBUS_FUNC_WRITE_SINGLE_REGISTER,
        MODBUS_FUNC_WRITE_MULTIPLE_COILS,
        MODBUS_FUNC_WRITE_MULTIPLE_REGISTERS,
    }
    WATCHED_FUNCTIONS = {
        MODBUS_FUNC_READ_COILS,
        MODBUS_FUNC_READ_DISCRETE_INPUTS,
        MODBUS_FUNC_READ_HOLDING_REGISTERS,
        MODBUS_FUNC_READ_INPUT_REGISTERS,
    }

    def __init__(self, logger: logging.Logger) -> None:
        self._coil_values: dict[int, bool] = {}
        self._discrete_input_values: dict[int, bool] = {}
        self._holding_register_values: dict[int, int] = {}
        self._input_register_values: dict[int, int] = {}
        self._run_task: asyncio.Task | None = None

        self._last_read_coils_request: dict[str, Any] | None = None
        self._last_read_discrete_inputs_request: dict[str, Any] | None = None
        self._last_read_holding_registers_request: dict[str, Any] | None = None
        self._last_read_input_registers_request: dict[str, Any] | None = None

        InternalModbusBaseClient.__init__(self, framer=Framer.RTU)
        logging_utils.LoggableMixin.__init__(self, "passive", logger)

    @abc.abstractmethod
    async def run(self) -> None:
        pass

    async def _run(self) -> None:
        try:
            await self.run()
        finally:
            self._run_task = None

    async def connect(self) -> None:
        if not self._run_task:
            self._run_task = asyncio.create_task(self._run())

    def close(self, reconnect: bool = False) -> None:
        if self._run_task:
            self._run_task.cancel()
            self._run_task = None

    @property
    def connected(self) -> bool:
        return bool(self._run_task)

    def get_coil_value(self, address: int) -> bool:
        return self._coil_values.get(address)

    def set_coil_value(self, address: int, value: bool) -> None:
        self._coil_values[address] = value

    def get_discrete_input_value(self, address: int) -> bool:
        return self._discrete_input_values.get(address)

    def set_discrete_input_value(self, address: int, value: bool) -> None:
        self._discrete_input_values[address] = value

    def get_input_register_value(self, address: int) -> int:
        return self._input_register_values.get(address)

    def set_input_register_value(self, address: int, value: int) -> None:
        self._input_register_values[address] = value

    def get_holding_register_value(self, address: int) -> int:
        return self._holding_register_values.get(address)

    def set_holding_register_value(self, address: int, value: int) -> None:
        self._holding_register_values[address] = value

    async def read_coils(
        self, address: int, count: int = 1, slave: int = 0, **kwargs
    ) -> bit_read_message.ReadCoilsResponse:
        return bit_read_message.ReadCoilsResponse(values=[self.get_coil_value(address + i) for i in range(count)])

    async def read_discrete_inputs(
        self, address: int, count: int = 1, slave: int = 0, **kwargs
    ) -> bit_read_message.ReadDiscreteInputsResponse:
        return bit_read_message.ReadDiscreteInputsResponse(
            values=[self.get_discrete_input_value(address + i) for i in range(count)]
        )

    async def read_input_registers(
        self, address: int, count: int = 1, slave: int = 0, **kwargs
    ) -> register_read_message.ReadInputRegistersResponse:
        return register_read_message.ReadInputRegistersResponse(
            values=[self.get_input_register_value(address + i) for i in range(count)]
        )

    async def read_holding_registers(
        self, address: int, count: int = 1, slave: int = 0, **kwargs
    ) -> register_read_message.ReadHoldingRegistersResponse:
        return register_read_message.ReadHoldingRegistersResponse(
            values=[self.get_holding_register_value(address + i) for i in range(count)]
        )

    async def write_coil(
        self, address: int, value: bool, slave: int = 0, **kwargs
    ) -> bit_write_message.WriteSingleCoilResponse:
        raise InternalPassiveException("Operation not supported in passive mode")

    async def write_coils(
        self, address: int, values: list[bool], slave: int = 0, **kwargs
    ) -> bit_write_message.WriteMultipleCoilsResponse:
        raise InternalPassiveException("Operation not supported in passive mode")

    async def write_register(
        self, address: int, value: int | float | str, slave: int = 0, **kwargs
    ) -> register_write_message.WriteSingleRegisterResponse:
        raise InternalPassiveException("Operation not supported in passive mode")

    async def write_registers(
        self,
        address: int,
        values: list[int | float | str],
        slave: int = 0,
        **kwargs,
    ) -> register_write_message.WriteMultipleRegistersResponse:
        raise InternalPassiveException("Operation not supported in passive mode")

    def process_read_coils_request(self, data: bytes) -> None:
        self._last_read_coils_request = {
            "address": int.from_bytes(data[:2], byteorder="big", signed=False),
            "count": int.from_bytes(data[2:4], byteorder="big", signed=False),
        }
        self.debug(
            "read coils request (address = %d, count = %d)",
            self._last_read_coils_request["address"],
            self._last_read_coils_request["count"],
        )

    def process_read_discrete_inputs_request(self, data: bytes) -> None:
        self._last_read_discrete_inputs_request = {
            "address": int.from_bytes(data[:2], byteorder="big", signed=False),
            "count": int.from_bytes(data[2:4], byteorder="big", signed=False),
        }
        self.debug(
            "read discrete inputs request (address = %d, count = %d)",
            self._last_read_discrete_inputs_request["address"],
            self._last_read_discrete_inputs_request["count"],
        )

    def process_read_holding_registers_request(self, data: bytes) -> None:
        self._last_read_holding_registers_request = {
            "address": int.from_bytes(data[:2], byteorder="big", signed=False),
            "count": int.from_bytes(data[2:4], byteorder="big", signed=False),
        }
        self.debug(
            "read holding registers request (address = %d, count = %d)",
            self._last_read_holding_registers_request["address"],
            self._last_read_holding_registers_request["count"],
        )

    def process_read_input_registers_request(self, data: bytes) -> None:
        self._last_read_input_registers_request = {
            "address": int.from_bytes(data[:2], byteorder="big", signed=False),
            "count": int.from_bytes(data[2:4], byteorder="big", signed=False),
        }
        self.debug(
            "read input registers request (address = %d, count = %d)",
            self._last_read_input_registers_request["address"],
            self._last_read_input_registers_request["count"],
        )

    def process_read_coils_response(self, data: bytes) -> None:
        if not self._last_read_coils_request:
            self.warning("ignoring read coils response without request")
            return

        if len(data) < 1:
            raise InvalidModbusFrame("Data too short (%s bytes)", len(data))
        byte_count = data[0]
        if len(data) != byte_count + 1:
            raise InvalidModbusFrame("Unexpected number of data bytes (%s != %s)", len(data), byte_count)
        expected_byte_count = math.ceil(self._last_read_coils_request["count"] / 8)
        if expected_byte_count != byte_count:
            self.warning("ignoring read coils response with different address count")
            return

        bits = int.from_bytes(data[1:], byteorder="little", signed=False)
        bit_string = f"{bits:b}"
        values = [b == "1" for b in reversed(bit_string)]
        for i, v in enumerate(values):
            self.set_coil_value(self._last_read_coils_request["address"] + i, v)

        self.debug("read coils response (values = [%s])", ", ".join(str(v).lower() for v in values))

        self._last_read_coils_request = None

    def process_read_discrete_inputs_response(self, data: bytes) -> None:
        if not self._last_read_discrete_inputs_request:
            self.warning("ignoring read discrete inputs response without request")
            return

        if len(data) < 1:
            raise InvalidModbusFrame("Data too short (%s bytes)", len(data))
        byte_count = data[0]
        if len(data) != byte_count + 1:
            raise InvalidModbusFrame("Unexpected number of data bytes (%s != %s)", len(data), byte_count)
        expected_byte_count = math.ceil(self._last_read_discrete_inputs_request["count"] / 8)
        if expected_byte_count != byte_count:
            self.warning("ignoring read discrete inputs response with different address count")
            return

        bits = int.from_bytes(data[1:], byteorder="little", signed=False)
        bit_string = f"{bits:b}"
        values = [b == "1" for b in reversed(bit_string)]
        for i, v in enumerate(values):
            self.set_discrete_input_value(self._last_read_discrete_inputs_request["address"] + i, v)

        self.debug("read discrete inputs response (values = [%s])", ", ".join(str(v).lower() for v in values))

        self._last_read_discrete_inputs_request = None

    def process_read_holding_registers_response(self, data: bytes) -> None:
        if not self._last_read_holding_registers_request:
            self.warning("ignoring read holding registers response without request")
            return

        if len(data) < 1:
            raise InvalidModbusFrame("Data too short (%s bytes)", len(data))
        byte_count = data[0]
        if len(data) != byte_count + 1:
            raise InvalidModbusFrame("Unexpected number of data bytes (%s != %s)", len(data), byte_count)
        expected_byte_count = self._last_read_holding_registers_request["count"] * 2
        if expected_byte_count != byte_count:
            self.warning("ignoring read holding registers response with different address count")
            return

        values = []
        for i in range(self._last_read_holding_registers_request["count"]):
            value = int.from_bytes(data[i * 2 + 1 : i * 2 + 3], byteorder="big", signed=False)
            values.append(value)
            self.set_holding_register_value(self._last_read_holding_registers_request["address"] + i, value)

        self.debug("read holding registers response (values = [%s])", ", ".join(f"0x{v:04X}" for v in values))

        self._last_read_holding_registers_request = None

    def process_read_input_registers_response(self, data: bytes) -> None:
        if not self._last_read_input_registers_request:
            self.warning("ignoring read input registers response without request")
            return

        if len(data) < 1:
            raise InvalidModbusFrame("Data too short (%s bytes)", len(data))
        byte_count = data[0]
        if len(data) != byte_count + 1:
            raise InvalidModbusFrame("Unexpected number of data bytes (%s != %s)", len(data), byte_count)
        expected_byte_count = self._last_read_input_registers_request["count"] * 2
        if expected_byte_count != byte_count:
            self.warning("ignoring read input registers response with different address count")
            return

        values = []
        for i in range(self._last_read_input_registers_request["count"]):
            value = int.from_bytes(data[i * 2 + 1 : i * 2 + 3], byteorder="big", signed=False)
            values.append(value)
            self.set_input_register_value(self._last_read_input_registers_request["address"] + i, value)

        self.debug("read input registers response (values = [%s])", ", ".join(f"0x{v:04X}" for v in values))

        self._last_read_input_registers_request = None

    @staticmethod
    def compute_crc(data: bytes) -> bytes:
        crc = 0xFFFF
        for b in data:
            crc = (crc ^ b) & 0xFFFF

            for i in range(8, 0, -1):
                if (crc & 0x0001) != 0:
                    crc >>= 1
                    crc ^= 0xA001
                else:
                    crc >>= 1

        return crc.to_bytes(2, byteorder="little", signed=False)



import logging

from qtoggleserver.peripherals import Peripheral

from .base import BaseModbus


class BaseModbusServer(BaseModbus, Peripheral):
    logger = logging.getLogger(__name__)

    def __init__(
        self,
        *,
        identity_vendor_name: str = "",
        identity_product_code: str = "",
        identity_major_minor_revision: str = "",
        identity_vendor_url: str = "",
        identity_product_name: str = "",
        identity_model_name: str = "",
        identity_user_application_name: str = "",
        **kwargs,
    ) -> None:
        self.identity_vendor_name: str = identity_vendor_name
        self.identity_product_code: str = identity_product_code
        self.identity_major_minor_revision: str = identity_major_minor_revision
        self.identity_vendor_url: str = identity_vendor_url
        self.identity_product_name: str = identity_product_name
        self.identity_model_name: str = identity_model_name
        self.identity_user_application_name: str = identity_user_application_name

        super().__init__(**kwargs)

    # async def make_port_args(self) -> list[Union[dict[str, Any], type[core_ports.BasePort]]]:
    #     port_args = []
    #     for id_, details in self.port_details.items():
    #         port_args.append({
    #             'driver': ...,
    #             'id': id_,
    #             **details
    #         })
    #
    #     return port_args

    # async def write_port_value(
    #     self,
    #     port: core_ports.BasePort,
    #     request_details: dict[str, Any],
    #     context: dict[str, Any]
    # ) -> None:
    # ...



import asyncio
import binascii
import logging

import serial

from .base import InternalPassiveClient, InternalPassiveException


class InvalidIpv4Packet(InternalPassiveException):
    pass


class InvalidModbusFrame(InternalPassiveException):
    pass


class InternalSerialClient(InternalPassiveClient):
    def __init__(
        self,
        *,
        serial_port: str,
        serial_baud: int,
        serial_stopbits: int,
        serial_bytesize: int,
        serial_parity: str,
        unit_id: int,
        timeout: int,
        logger: logging.Logger,
    ) -> None:
        self.serial_port: str = serial_port
        self.serial_baud: int = serial_baud
        self.serial_stopbits: int = serial_stopbits
        self.serial_bytesize: int = serial_bytesize
        self.serial_parity: str = serial_parity
        self.unit_id: int = unit_id
        self.timeout: int = timeout

        super().__init__(logger)
        self.set_logger_name("passive_serial")

    def make_serial(self) -> serial.Serial:
        return serial.Serial(
            port=self.serial_port,
            baudrate=self.serial_baud,
            stopbits=self.serial_stopbits,
            bytesize=self.serial_bytesize,
            parity=self.serial_parity,
            timeout=self.timeout,
            write_timeout=self.timeout,
        )

    async def run(self) -> None:
        while True:
            try:
                ser = self.make_serial()
                buffer = b""
                for _ in range(self.timeout * 10):
                    buffer += ser.read_all()
                    await asyncio.sleep(0.1)
                if buffer:
                    self.debug("sniffed %d bytes" % len(buffer))
                    while True:
                        consumed_bytes = self._process_modbus_data(buffer)
                        if not consumed_bytes:
                            break
                        buffer = buffer[consumed_bytes:]

            except asyncio.CancelledError:
                self.debug("task stopped")
                break

    def _process_modbus_data(self, buffer: bytes) -> int:
        if len(buffer) < 5:
            return 0

        unit_id = buffer[0]
        if unit_id != self.unit_id:
            return 0

        function = buffer[1]
        if function not in self.WATCHED_FUNCTIONS:
            return 0

        # Try to see if we've got a response
        byte_count = buffer[2]
        if len(buffer) >= byte_count + 5:
            crc = buffer[byte_count + 3 : byte_count + 5]
            if self.compute_crc(buffer[: byte_count + 3]) == crc:
                frame = buffer[: byte_count + 5]
                self._process_modbus_response(frame)
                return len(frame)

        # Try to see if we've got a request
        if len(buffer) >= 8:
            crc = buffer[6:8]
            if self.compute_crc(buffer[:6]) == crc:
                frame = buffer[0:8]
                self._process_modbus_request(frame)
                return len(frame)

        return 0

    def _process_modbus_request(self, frame: bytes) -> None:
        self.debug("got Modbus request: %s", binascii.hexlify(frame).decode())

        function = frame[1]
        data = frame[2:-2]
        if function == self.MODBUS_FUNC_READ_COILS:
            self.process_read_coils_request(data)
        elif function == self.MODBUS_FUNC_READ_DISCRETE_INPUTS:
            self.process_read_discrete_inputs_request(data)
        elif function == self.MODBUS_FUNC_READ_HOLDING_REGISTERS:
            self.process_read_holding_registers_request(data)
        elif function == self.MODBUS_FUNC_READ_INPUT_REGISTERS:
            self.process_read_input_registers_request(data)

    def _process_modbus_response(self, frame: bytes) -> None:
        self.debug("got Modbus response: %s", binascii.hexlify(frame).decode())

        function = frame[1]
        data = frame[2:-2]
        if function == self.MODBUS_FUNC_READ_COILS:
            self.process_read_coils_response(data)
        elif function == self.MODBUS_FUNC_READ_DISCRETE_INPUTS:
            self.process_read_discrete_inputs_response(data)
        elif function == self.MODBUS_FUNC_READ_HOLDING_REGISTERS:
            self.process_read_holding_registers_response(data)
        elif function == self.MODBUS_FUNC_READ_INPUT_REGISTERS:
            self.process_read_input_registers_response(data)
